package com.sun.jndi.ldap;

import javax.naming.ldap.UnsolicitedNotificationListener;
import javax.naming.event.NamingListener;
import com.sun.jndi.ldap.ext.StartTlsResponseImpl;
import javax.naming.ldap.ExtendedResponse;
import javax.naming.ldap.ExtendedRequest;
import javax.naming.directory.InvalidAttributeIdentifierException;
import javax.naming.ServiceUnavailableException;
import javax.naming.TimeLimitExceededException;
import javax.naming.SizeLimitExceededException;
import javax.naming.ContextNotEmptyException;
import javax.naming.directory.NoSuchAttributeException;
import javax.naming.directory.InvalidAttributeValueException;
import javax.naming.NoPermissionException;
import javax.naming.directory.InvalidSearchFilterException;
import javax.naming.AuthenticationException;
import javax.naming.NameAlreadyBoundException;
import javax.naming.AuthenticationNotSupportedException;
import javax.naming.directory.AttributeInUseException;
import java.util.Iterator;
import java.util.Collections;
import java.util.stream.Collectors;
import java.util.Arrays;
import java.security.PrivilegedAction;
import java.security.AccessController;
import java.util.Locale;
import java.util.StringTokenizer;
import com.sun.jndi.toolkit.dir.SearchFilter;
import javax.naming.ConfigurationException;
import javax.naming.OperationNotSupportedException;
import javax.naming.directory.SearchResult;
import javax.naming.directory.SchemaViolationException;
import javax.naming.directory.ModificationItem;
import javax.naming.Binding;
import javax.naming.LimitExceededException;
import javax.naming.NameClassPair;
import javax.naming.spi.DirectoryManager;
import javax.naming.directory.SearchControls;
import javax.naming.NamingEnumeration;
import java.util.List;
import javax.naming.ldap.Rdn;
import javax.naming.ldap.LdapName;
import javax.naming.directory.BasicAttributes;
import javax.naming.directory.BasicAttribute;
import javax.naming.Context;
import javax.naming.PartialResultException;
import javax.naming.InvalidNameException;
import javax.naming.CompositeName;
import javax.naming.directory.Attribute;
import javax.naming.NameNotFoundException;
import java.io.IOException;
import javax.naming.CommunicationException;
import javax.naming.directory.Attributes;
import com.sun.jndi.toolkit.ctx.Continuation;
import javax.naming.NamingException;
import javax.naming.directory.DirContext;
import java.io.OutputStream;
import java.util.Vector;
import javax.naming.Name;
import java.util.Hashtable;
import com.sun.jndi.toolkit.dir.HierMemDirCtx;
import javax.naming.ldap.Control;
import javax.naming.ldap.ControlFactory;
import javax.naming.NameParser;
import java.util.Set;
import javax.naming.ldap.LdapContext;
import javax.naming.event.EventDirContext;
import com.sun.jndi.toolkit.ctx.ComponentDirContext;

public final class LdapCtx extends ComponentDirContext implements EventDirContext, LdapContext
{
    private static final boolean debug = false;
    private static final boolean HARD_CLOSE = true;
    private static final boolean SOFT_CLOSE = false;
    public static final int DEFAULT_PORT = 389;
    public static final int DEFAULT_SSL_PORT = 636;
    public static final String DEFAULT_HOST = "localhost";
    private static final boolean DEFAULT_DELETE_RDN = true;
    private static final boolean DEFAULT_TYPES_ONLY = false;
    private static final int DEFAULT_DEREF_ALIASES = 3;
    private static final int DEFAULT_LDAP_VERSION = 32;
    private static final int DEFAULT_BATCH_SIZE = 1;
    private static final int DEFAULT_REFERRAL_MODE = 3;
    private static final char DEFAULT_REF_SEPARATOR = '#';
    static final String DEFAULT_SSL_FACTORY = "javax.net.ssl.SSLSocketFactory";
    private static final int DEFAULT_REFERRAL_LIMIT = 10;
    private static final String STARTTLS_REQ_OID = "1.3.6.1.4.1.1466.20037";
    private static final String[] SCHEMA_ATTRIBUTES;
    private static final String VERSION = "java.naming.ldap.version";
    private static final String BINARY_ATTRIBUTES = "java.naming.ldap.attributes.binary";
    private static final String DELETE_RDN = "java.naming.ldap.deleteRDN";
    private static final String DEREF_ALIASES = "java.naming.ldap.derefAliases";
    private static final String TYPES_ONLY = "java.naming.ldap.typesOnly";
    private static final String REF_SEPARATOR = "java.naming.ldap.ref.separator";
    private static final String SOCKET_FACTORY = "java.naming.ldap.factory.socket";
    static final String BIND_CONTROLS = "java.naming.ldap.control.connect";
    private static final String REFERRAL_LIMIT = "java.naming.ldap.referral.limit";
    private static final String TRACE_BER = "com.sun.jndi.ldap.trace.ber";
    private static final String NETSCAPE_SCHEMA_BUG = "com.sun.jndi.ldap.netscape.schemaBugs";
    private static final String OLD_NETSCAPE_SCHEMA_BUG = "com.sun.naming.netscape.schemaBugs";
    private static final String CONNECT_TIMEOUT = "com.sun.jndi.ldap.connect.timeout";
    private static final String READ_TIMEOUT = "com.sun.jndi.ldap.read.timeout";
    private static final String ENABLE_POOL = "com.sun.jndi.ldap.connect.pool";
    private static final String DOMAIN_NAME = "com.sun.jndi.ldap.domainname";
    private static final String WAIT_FOR_REPLY = "com.sun.jndi.ldap.search.waitForReply";
    private static final String REPLY_QUEUE_SIZE = "com.sun.jndi.ldap.search.replyQueueSize";
    private static final String ALLOWED_MECHS_SP = "jdk.jndi.ldap.mechsAllowedToSendCredentials";
    private static final String ALLOWED_MECHS_SP_VALUE;
    private static final Set<String> MECHS_ALLOWED_BY_SP;
    private static final String UNSECURED_CRED_TRANSMIT_MSG = "Transmission of credentials over unsecured connection is not allowed";
    private static final NameParser parser;
    private static final ControlFactory myResponseControlFactory;
    private static final Control manageReferralControl;
    private static final HierMemDirCtx EMPTY_SCHEMA;
    int port_number;
    String hostname;
    LdapClient clnt;
    Hashtable<String, Object> envprops;
    int handleReferrals;
    boolean hasLdapsScheme;
    String currentDN;
    Name currentParsedDN;
    Vector<Control> respCtls;
    Control[] reqCtls;
    volatile boolean contextSeenStartTlsEnabled;
    private OutputStream trace;
    private boolean netscapeSchemaBug;
    private Control[] bindCtls;
    private int referralHopLimit;
    private Hashtable<String, DirContext> schemaTrees;
    private int batchSize;
    private boolean deleteRDN;
    private boolean typesOnly;
    private int derefAliases;
    private char addrEncodingSeparator;
    private Hashtable<String, Boolean> binaryAttrs;
    private int connectTimeout;
    private int readTimeout;
    private boolean waitForReply;
    private int replyQueueSize;
    private boolean useSsl;
    private boolean useDefaultPortNumber;
    private boolean parentIsLdapCtx;
    private int hopCount;
    private String url;
    private EventSupport eventSupport;
    private boolean unsolicited;
    private boolean sharable;
    private int enumCount;
    private boolean closeRequested;
    
    public LdapCtx(String s, String s2, int port_number, Hashtable<?, ?> hashtable, boolean b) throws NamingException {
        this.hostname = null;
        this.clnt = null;
        this.envprops = null;
        this.handleReferrals = 3;
        this.hasLdapsScheme = false;
        this.respCtls = null;
        this.reqCtls = null;
        this.trace = null;
        this.netscapeSchemaBug = false;
        this.bindCtls = null;
        this.referralHopLimit = 10;
        this.schemaTrees = null;
        this.batchSize = 1;
        this.deleteRDN = true;
        this.typesOnly = false;
        this.derefAliases = 3;
        this.addrEncodingSeparator = '#';
        this.binaryAttrs = null;
        this.connectTimeout = -1;
        this.readTimeout = -1;
        this.waitForReply = true;
        this.replyQueueSize = -1;
        this.useSsl = false;
        this.useDefaultPortNumber = false;
        this.parentIsLdapCtx = false;
        this.hopCount = 1;
        this.url = null;
        this.unsolicited = false;
        this.sharable = true;
        this.enumCount = 0;
        this.closeRequested = false;
        this.hasLdapsScheme = b;
        this.useSsl = b;
        if (hashtable != null) {
            this.envprops = (Hashtable)hashtable.clone();
            if ("ssl".equals(this.envprops.get("java.naming.security.protocol"))) {
                this.useSsl = true;
            }
            this.trace = (OutputStream)this.envprops.get("com.sun.jndi.ldap.trace.ber");
            if (hashtable.get("com.sun.jndi.ldap.netscape.schemaBugs") != null || hashtable.get("com.sun.naming.netscape.schemaBugs") != null) {
                this.netscapeSchemaBug = true;
            }
        }
        this.currentDN = ((s != null) ? s : "");
        this.currentParsedDN = LdapCtx.parser.parse(this.currentDN);
        this.hostname = ((s2 != null && s2.length() > 0) ? s2 : "localhost");
        if (this.hostname.charAt(0) == '[') {
            this.hostname = this.hostname.substring(1, this.hostname.length() - 1);
        }
        if (port_number > 0) {
            this.port_number = port_number;
        }
        else {
            this.port_number = (this.useSsl ? 636 : 389);
            this.useDefaultPortNumber = true;
        }
        this.schemaTrees = new Hashtable(11, 0.75f);
        this.initEnv();
        try {
            this.connect(false);
        }
        catch (NamingException ex) {
            try {
                this.close();
            }
            catch (Exception ex2) {}
            throw ex;
        }
    }
    
    LdapCtx(LdapCtx ldapCtx, String currentDN) throws NamingException {
        this.hostname = null;
        this.clnt = null;
        this.envprops = null;
        this.handleReferrals = 3;
        this.hasLdapsScheme = false;
        this.respCtls = null;
        this.reqCtls = null;
        this.trace = null;
        this.netscapeSchemaBug = false;
        this.bindCtls = null;
        this.referralHopLimit = 10;
        this.schemaTrees = null;
        this.batchSize = 1;
        this.deleteRDN = true;
        this.typesOnly = false;
        this.derefAliases = 3;
        this.addrEncodingSeparator = '#';
        this.binaryAttrs = null;
        this.connectTimeout = -1;
        this.readTimeout = -1;
        this.waitForReply = true;
        this.replyQueueSize = -1;
        this.useSsl = false;
        this.useDefaultPortNumber = false;
        this.parentIsLdapCtx = false;
        this.hopCount = 1;
        this.url = null;
        this.unsolicited = false;
        this.sharable = true;
        this.enumCount = 0;
        this.closeRequested = false;
        this.useSsl = ldapCtx.useSsl;
        this.hasLdapsScheme = ldapCtx.hasLdapsScheme;
        this.useDefaultPortNumber = ldapCtx.useDefaultPortNumber;
        this.hostname = ldapCtx.hostname;
        this.port_number = ldapCtx.port_number;
        this.currentDN = currentDN;
        if (ldapCtx.currentDN == this.currentDN) {
            this.currentParsedDN = ldapCtx.currentParsedDN;
        }
        else {
            this.currentParsedDN = LdapCtx.parser.parse(this.currentDN);
        }
        this.envprops = ldapCtx.envprops;
        this.schemaTrees = ldapCtx.schemaTrees;
        (this.clnt = ldapCtx.clnt).incRefCount();
        this.parentIsLdapCtx = ((currentDN != null && !currentDN.equals(ldapCtx.currentDN)) || ldapCtx.parentIsLdapCtx);
        this.trace = ldapCtx.trace;
        this.netscapeSchemaBug = ldapCtx.netscapeSchemaBug;
        this.initEnv();
    }
    
    @Override
    public LdapContext newInstance(Control[] requestControls) throws NamingException {
        LdapCtx ldapCtx = new LdapCtx(this, this.currentDN);
        ldapCtx.setRequestControls(requestControls);
        return ldapCtx;
    }
    
    @Override
    protected void c_bind(Name name, Object o, Continuation continuation) throws NamingException {
        this.c_bind(name, o, null, continuation);
    }
    
    @Override
    protected void c_bind(Name remain, Object o, Attributes attributes, Continuation continuation) throws NamingException {
        continuation.setError(this, remain);
        try {
            this.ensureOpen();
            if (o == null) {
                if (attributes == null) {
                    throw new IllegalArgumentException("cannot bind null object with no attributes");
                }
            }
            else {
                attributes = Obj.determineBindAttrs(this.addrEncodingSeparator, o, attributes, false, remain, this, this.envprops);
            }
            String fullyQualifiedName = this.fullyQualifiedName(remain);
            attributes = addRdnAttributes(fullyQualifiedName, attributes, attributes != attributes);
            LdapResult add = this.clnt.add(new LdapEntry(fullyQualifiedName, attributes), this.reqCtls);
            this.respCtls = add.resControls;
            if (add.status != 0) {
                this.processReturnCode(add, remain);
            }
        }
        catch (LdapReferralException e) {
            if (this.handleReferrals == 2) {
                throw continuation.fillInException(e);
            }
            while (true) {
                LdapReferralContext ldapReferralContext = (LdapReferralContext)e.getReferralContext(this.envprops, this.bindCtls);
                try {
                    ldapReferralContext.bind(remain, o, attributes);
                    return;
                }
                catch (LdapReferralException ex) {
                    e = ex;
                }
                finally {
                    ldapReferralContext.close();
                }
            }
        }
        catch (IOException rootCause) {
            CommunicationException e2 = new CommunicationException(rootCause.getMessage());
            e2.setRootCause(rootCause);
            throw continuation.fillInException(e2);
        }
        catch (NamingException e3) {
            throw continuation.fillInException(e3);
        }
    }
    
    @Override
    protected void c_rebind(Name name, Object o, Continuation continuation) throws NamingException {
        this.c_rebind(name, o, null, continuation);
    }
    
    @Override
    protected void c_rebind(Name remain, Object o, Attributes attributes, Continuation continuation) throws NamingException {
        continuation.setError(this, remain);
        try {
            Attributes c_getAttributes = null;
            try {
                c_getAttributes = this.c_getAttributes(remain, null, continuation);
            }
            catch (NameNotFoundException ex) {}
            if (c_getAttributes == null) {
                this.c_bind(remain, o, attributes, continuation);
                return;
            }
            if (attributes == null && o instanceof DirContext) {
                attributes = ((DirContext)o).getAttributes("");
            }
            Attributes attributes2 = (Attributes)c_getAttributes.clone();
            if (attributes == null) {
                Attribute value = c_getAttributes.get(Obj.JAVA_ATTRIBUTES[0]);
                if (value != null) {
                    Attribute attribute = (Attribute)value.clone();
                    for (int i = 0; i < Obj.JAVA_OBJECT_CLASSES.length; ++i) {
                        attribute.remove(Obj.JAVA_OBJECT_CLASSES_LOWER[i]);
                        attribute.remove(Obj.JAVA_OBJECT_CLASSES[i]);
                    }
                    c_getAttributes.put(attribute);
                }
                for (int j = 1; j < Obj.JAVA_ATTRIBUTES.length; ++j) {
                    c_getAttributes.remove(Obj.JAVA_ATTRIBUTES[j]);
                }
                attributes = c_getAttributes;
            }
            if (o != null) {
                attributes = Obj.determineBindAttrs(this.addrEncodingSeparator, o, attributes, attributes != attributes, remain, this, this.envprops);
            }
            String fullyQualifiedName = this.fullyQualifiedName(remain);
            LdapResult ldapResult = this.clnt.delete(fullyQualifiedName, this.reqCtls);
            this.respCtls = ldapResult.resControls;
            if (ldapResult.status != 0) {
                this.processReturnCode(ldapResult, remain);
                return;
            }
            NamingException | IOException ex2 = null;
            try {
                attributes = addRdnAttributes(fullyQualifiedName, attributes, attributes != attributes);
                ldapResult = this.clnt.add(new LdapEntry(fullyQualifiedName, attributes), this.reqCtls);
                if (ldapResult.resControls != null) {
                    this.respCtls = appendVector(this.respCtls, ldapResult.resControls);
                }
            }
            catch (NamingException | IOException ex3) {
                ex2 = ex3;
            }
            if ((ex2 != null && !(ex2 instanceof LdapReferralException)) || ldapResult.status != 0) {
                LdapResult add = this.clnt.add(new LdapEntry(fullyQualifiedName, attributes2), this.reqCtls);
                if (add.resControls != null) {
                    this.respCtls = appendVector(this.respCtls, add.resControls);
                }
                if (ex2 == null) {
                    this.processReturnCode(ldapResult, remain);
                }
            }
            if ((ex2 instanceof NamingException)) {
                throw (NamingException)ex2;
            }
            if ((ex2 instanceof IOException)) {
                throw (IOException)ex2;
            }
        }
        catch (LdapReferralException e) {
            if (this.handleReferrals == 2) {
                throw continuation.fillInException(e);
            }
            while (true) {
                LdapReferralContext ldapReferralContext = (LdapReferralContext)e.getReferralContext(this.envprops, this.bindCtls);
                try {
                    ldapReferralContext.rebind(remain, o, attributes);
                    return;
                }
                catch (LdapReferralException ex4) {
                    e = ex4;
                }
                finally {
                    ldapReferralContext.close();
                }
            }
        }
        catch (IOException rootCause) {
            CommunicationException e2 = new CommunicationException(rootCause.getMessage());
            e2.setRootCause(rootCause);
            throw continuation.fillInException(e2);
        }
        catch (NamingException e3) {
            throw continuation.fillInException(e3);
        }
    }
    
    @Override
    protected void c_unbind(Name remain, Continuation continuation) throws NamingException {
        continuation.setError(this, remain);
        try {
            this.ensureOpen();
            String fullyQualifiedName = this.fullyQualifiedName(remain);
            LdapResult delete = this.clnt.delete(fullyQualifiedName, this.reqCtls);
            this.respCtls = delete.resControls;
            this.adjustDeleteStatus(fullyQualifiedName, delete);
            if (delete.status != 0) {
                this.processReturnCode(delete, remain);
            }
        }
        catch (LdapReferralException e) {
            if (this.handleReferrals == 2) {
                throw continuation.fillInException(e);
            }
            while (true) {
                LdapReferralContext ldapReferralContext = (LdapReferralContext)e.getReferralContext(this.envprops, this.bindCtls);
                try {
                    ldapReferralContext.unbind(remain);
                    return;
                }
                catch (LdapReferralException ex) {
                    e = ex;
                }
                finally {
                    ldapReferralContext.close();
                }
            }
        }
        catch (IOException rootCause) {
            CommunicationException e2 = new CommunicationException(rootCause.getMessage());
            e2.setRootCause(rootCause);
            throw continuation.fillInException(e2);
        }
        catch (NamingException e3) {
            throw continuation.fillInException(e3);
        }
    }
    
    @Override
    protected void c_rename(Name remain, Name name, Continuation continuation) throws NamingException {
        String value = null;
        String fullyQualifiedName = null;
        continuation.setError(this, remain);
        try {
            this.ensureOpen();
            Name name2;
            if (remain.isEmpty()) {
                name2 = LdapCtx.parser.parse("");
            }
            else {
                Name parse = LdapCtx.parser.parse(remain.get(0));
                name2 = parse.getPrefix(parse.size() - 1);
            }
            Name parse2;
            if ((name instanceof CompositeName)) {
                parse2 = LdapCtx.parser.parse(name.get(0));
            }
            else {
                parse2 = name;
            }
            Name prefix = parse2.getPrefix(parse2.size() - 1);
            if (!name2.equals(prefix)) {
                if (!this.clnt.isLdapv3) {
                    throw new InvalidNameException("LDAPv2 doesn't support changing the parent as a result of a rename");
                }
                fullyQualifiedName = this.fullyQualifiedName(prefix.toString());
            }
            value = parse2.get(parse2.size() - 1);
            LdapResult moddn = this.clnt.moddn(this.fullyQualifiedName(remain), value, this.deleteRDN, fullyQualifiedName, this.reqCtls);
            this.respCtls = moddn.resControls;
            if (moddn.status != 0) {
                this.processReturnCode(moddn, remain);
            }
        }
        catch (LdapReferralException ex) {
            ex.setNewRdn(value);
            if (fullyQualifiedName != null) {
                PartialResultException e = new PartialResultException("Cannot continue referral processing when newSuperior is nonempty: " + fullyQualifiedName);
                e.setRootCause(continuation.fillInException(ex));
                throw continuation.fillInException(e);
            }
            if (this.handleReferrals == 2) {
                throw continuation.fillInException(ex);
            }
            while (true) {
                LdapReferralContext ldapReferralContext = (LdapReferralContext)ex.getReferralContext(this.envprops, this.bindCtls);
                try {
                    ldapReferralContext.rename(remain, name);
                    return;
                }
                catch (LdapReferralException ex2) {
                    ex = ex2;
                }
                finally {
                    ldapReferralContext.close();
                }
            }
        }
        catch (IOException rootCause) {
            CommunicationException e2 = new CommunicationException(rootCause.getMessage());
            e2.setRootCause(rootCause);
            throw continuation.fillInException(e2);
        }
        catch (NamingException e3) {
            throw continuation.fillInException(e3);
        }
    }
    
    @Override
    protected Context c_createSubcontext(Name name, Continuation continuation) throws NamingException {
        return this.c_createSubcontext(name, null, continuation);
    }
    
    @Override
    protected DirContext c_createSubcontext(Name remain, Attributes addRdnAttributes, Continuation continuation) throws NamingException {
        continuation.setError(this, remain);
        try {
            this.ensureOpen();
            if (addRdnAttributes == null) {
                BasicAttribute basicAttribute = new BasicAttribute(Obj.JAVA_ATTRIBUTES[0], Obj.JAVA_OBJECT_CLASSES[0]);
                basicAttribute.add("top");
                addRdnAttributes = new BasicAttributes(true);
                addRdnAttributes.put(basicAttribute);
            }
            String fullyQualifiedName = this.fullyQualifiedName(remain);
            addRdnAttributes = addRdnAttributes(fullyQualifiedName, addRdnAttributes, addRdnAttributes != addRdnAttributes);
            LdapResult add = this.clnt.add(new LdapEntry(fullyQualifiedName, addRdnAttributes), this.reqCtls);
            this.respCtls = add.resControls;
            if (add.status != 0) {
                this.processReturnCode(add, remain);
                return null;
            }
            return new LdapCtx(this, fullyQualifiedName);
        }
        catch (LdapReferralException e) {
            if (this.handleReferrals == 2) {
                throw continuation.fillInException(e);
            }
            while (true) {
                LdapReferralContext ldapReferralContext = (LdapReferralContext)e.getReferralContext(this.envprops, this.bindCtls);
                try {
                    return ldapReferralContext.createSubcontext(remain, addRdnAttributes);
                }
                catch (LdapReferralException ex) {
                    e = ex;
                }
                finally {
                    ldapReferralContext.close();
                }
            }
        }
        catch (IOException rootCause) {
            CommunicationException e2 = new CommunicationException(rootCause.getMessage());
            e2.setRootCause(rootCause);
            throw continuation.fillInException(e2);
        }
        catch (NamingException e3) {
            throw continuation.fillInException(e3);
        }
    }
    
    @Override
    protected void c_destroySubcontext(Name remain, Continuation continuation) throws NamingException {
        continuation.setError(this, remain);
        try {
            this.ensureOpen();
            String fullyQualifiedName = this.fullyQualifiedName(remain);
            LdapResult delete = this.clnt.delete(fullyQualifiedName, this.reqCtls);
            this.respCtls = delete.resControls;
            this.adjustDeleteStatus(fullyQualifiedName, delete);
            if (delete.status != 0) {
                this.processReturnCode(delete, remain);
            }
        }
        catch (LdapReferralException e) {
            if (this.handleReferrals == 2) {
                throw continuation.fillInException(e);
            }
            while (true) {
                LdapReferralContext ldapReferralContext = (LdapReferralContext)e.getReferralContext(this.envprops, this.bindCtls);
                try {
                    ldapReferralContext.destroySubcontext(remain);
                    return;
                }
                catch (LdapReferralException ex) {
                    e = ex;
                }
                finally {
                    ldapReferralContext.close();
                }
            }
        }
        catch (IOException rootCause) {
            CommunicationException e2 = new CommunicationException(rootCause.getMessage());
            e2.setRootCause(rootCause);
            throw continuation.fillInException(e2);
        }
        catch (NamingException e3) {
            throw continuation.fillInException(e3);
        }
    }
    
    private static Attributes addRdnAttributes(String name, Attributes attributes, boolean b) throws NamingException {
        if (name.equals("")) {
            return attributes;
        }
        List<Rdn> rdns = new LdapName(name).getRdns();
        NamingEnumeration<Object> all = ((Rdn)rdns.get(rdns.size() - 1)).toAttributes().getAll();
        while (all.hasMore()) {
            Attribute attribute = (Attribute)all.next();
            if (attributes.get(attribute.getID()) != null) {
                continue;
            }
            if (!attributes.isCaseIgnored() && containsIgnoreCase(attributes.getIDs(), attribute.getID())) {
                continue;
            }
            if (!b) {
                attributes = (Attributes)attributes.clone();
                b = true;
            }
            attributes.put(attribute);
        }
        return attributes;
    }
    
    private static boolean containsIgnoreCase(NamingEnumeration<String> namingEnumeration, String anotherString) throws NamingException {
        while (namingEnumeration.hasMore()) {
            if (!((String)namingEnumeration.next()).equalsIgnoreCase(anotherString)) {
                continue;
            }
            return true;
        }
        return false;
    }
    
    private void adjustDeleteStatus(String s, LdapResult ldapResult) {
        if (ldapResult.status == 32 && ldapResult.matchedDN != null) {
            try {
                if (LdapCtx.parser.parse(s).size() - LdapCtx.parser.parse(ldapResult.matchedDN).size() == 1) {
                    ldapResult.status = 0;
                }
            }
            catch (NamingException ex) {}
        }
    }
    
    private static <T> Vector<T> appendVector(Vector<T> vector, Vector<T> vector2) {
        if (vector == null) {
            vector = vector2;
            return vector;
        }
        for (int i = 0; i < vector2.size(); ++i) {
            vector.addElement(vector2.elementAt(i));
        }
        return vector;
    }
    
    @Override
    protected Object c_lookupLink(Name name, Continuation continuation) throws NamingException {
        return this.c_lookup(name, continuation);
    }
    
    @Override
    protected Object c_lookup(Name name, Continuation continuation) throws NamingException {
        continuation.setError(this, name);
        Object decodeObject = null;
        BasicAttributes attributes;
        try {
            SearchControls searchControls = new SearchControls();
            searchControls.setSearchScope(0);
            searchControls.setReturningAttributes((String[])null);
            searchControls.setReturningObjFlag(true);
            LdapResult doSearchOnce = this.doSearchOnce(name, "(objectClass=*)", searchControls, true);
            this.respCtls = doSearchOnce.resControls;
            if (doSearchOnce.status != 0) {
                this.processReturnCode(doSearchOnce, name);
            }
            if (doSearchOnce.entries == null || doSearchOnce.entries.size() != 1) {
                attributes = new BasicAttributes(true);
            }
            else {
                LdapEntry ldapEntry = (LdapEntry)doSearchOnce.entries.elementAt(0);
                attributes = ldapEntry.attributes;
                Vector<Control> respCtls = ldapEntry.respCtls;
                if (respCtls != null) {
                    appendVector(this.respCtls, respCtls);
                }
            }
            if (attributes.get(Obj.JAVA_ATTRIBUTES[2]) != null) {
                decodeObject = Obj.decodeObject(attributes);
            }
            if (decodeObject == null) {
                decodeObject = new LdapCtx(this, this.fullyQualifiedName(name));
            }
        }
        catch (LdapReferralException e) {
            if (this.handleReferrals == 2) {
                throw continuation.fillInException(e);
            }
            while (true) {
                LdapReferralContext ldapReferralContext = (LdapReferralContext)e.getReferralContext(this.envprops, this.bindCtls);
                try {
                    return ldapReferralContext.lookup(name);
                }
                catch (LdapReferralException ex) {
                    e = ex;
                }
                finally {
                    ldapReferralContext.close();
                }
            }
        }
        catch (NamingException e2) {
            throw continuation.fillInException(e2);
        }
        try {
            return DirectoryManager.getObjectInstance(decodeObject, name, this, this.envprops, attributes);
        }
        catch (NamingException e3) {
            throw continuation.fillInException(e3);
        }
        catch (Exception rootCause) {
            NamingException e4 = new NamingException("problem generating object using object factory");
            e4.setRootCause(rootCause);
            throw continuation.fillInException(e4);
        }
    }
    
    @Override
    protected NamingEnumeration<NameClassPair> c_list(Name remain, Continuation continuation) throws NamingException {
        SearchControls searchControls = new SearchControls();
        searchControls.setReturningAttributes(new String[] { Obj.JAVA_ATTRIBUTES[0], Obj.JAVA_ATTRIBUTES[2] });
        searchControls.setReturningObjFlag(true);
        continuation.setError(this, remain);
        LdapResult doSearch = null;
        try {
            doSearch = this.doSearch(remain, "(objectClass=*)", searchControls, true, true);
            if (doSearch.status != 0 || doSearch.referrals != null) {
                this.processReturnCode(doSearch, remain);
            }
            return new LdapNamingEnumeration(this, doSearch, remain, continuation);
        }
        catch (LdapReferralException e) {
            if (this.handleReferrals == 2) {
                throw continuation.fillInException(e);
            }
            while (true) {
                LdapReferralContext ldapReferralContext = (LdapReferralContext)e.getReferralContext(this.envprops, this.bindCtls);
                try {
                    return ldapReferralContext.list(remain);
                }
                catch (LdapReferralException ex) {
                    e = ex;
                }
                finally {
                    ldapReferralContext.close();
                }
            }
        }
        catch (LimitExceededException e2) {
            LdapNamingEnumeration ldapNamingEnumeration = new LdapNamingEnumeration(this, doSearch, remain, continuation);
            ldapNamingEnumeration.setNamingException((NamingException)continuation.fillInException(e2));
            return ldapNamingEnumeration;
        }
        catch (PartialResultException e3) {
            LdapNamingEnumeration ldapNamingEnumeration2 = new LdapNamingEnumeration(this, doSearch, remain, continuation);
            ldapNamingEnumeration2.setNamingException((NamingException)continuation.fillInException(e3));
            return ldapNamingEnumeration2;
        }
        catch (NamingException e4) {
            throw continuation.fillInException(e4);
        }
    }
    
    @Override
    protected NamingEnumeration<Binding> c_listBindings(Name remain, Continuation continuation) throws NamingException {
        SearchControls searchControls = new SearchControls();
        searchControls.setReturningAttributes((String[])null);
        searchControls.setReturningObjFlag(true);
        continuation.setError(this, remain);
        LdapResult doSearch = null;
        try {
            doSearch = this.doSearch(remain, "(objectClass=*)", searchControls, true, true);
            if (doSearch.status != 0 || doSearch.referrals != null) {
                this.processReturnCode(doSearch, remain);
            }
            return new LdapBindingEnumeration(this, doSearch, remain, continuation);
        }
        catch (LdapReferralException e) {
            if (this.handleReferrals == 2) {
                throw continuation.fillInException(e);
            }
            while (true) {
                LdapReferralContext ldapReferralContext = (LdapReferralContext)e.getReferralContext(this.envprops, this.bindCtls);
                try {
                    return ldapReferralContext.listBindings(remain);
                }
                catch (LdapReferralException ex) {
                    e = ex;
                }
                finally {
                    ldapReferralContext.close();
                }
            }
        }
        catch (LimitExceededException e2) {
            LdapBindingEnumeration ldapBindingEnumeration = new LdapBindingEnumeration(this, doSearch, remain, continuation);
            ldapBindingEnumeration.setNamingException(continuation.fillInException(e2));
            return ldapBindingEnumeration;
        }
        catch (PartialResultException e3) {
            LdapBindingEnumeration ldapBindingEnumeration2 = new LdapBindingEnumeration(this, doSearch, remain, continuation);
            ldapBindingEnumeration2.setNamingException(continuation.fillInException(e3));
            return ldapBindingEnumeration2;
        }
        catch (NamingException e4) {
            throw continuation.fillInException(e4);
        }
    }
    
    @Override
    protected NameParser c_getNameParser(Name name, Continuation continuation) throws NamingException {
        continuation.setSuccess();
        return LdapCtx.parser;
    }
    
    @Override
    public String getNameInNamespace() {
        return this.currentDN;
    }
    
    @Override
    public Name composeName(Name add, Name add2) throws NamingException {
        if (add instanceof LdapName && add2 instanceof LdapName) {
            Name name2 = (Name)(Name)add2.clone();
            name2.addAll(add);
            return new CompositeName().add(name2.toString());
        }
        if (!(add instanceof CompositeName)) {
            add = new CompositeName().add(add.toString());
        }
        if (!(add2 instanceof CompositeName)) {
            add2 = new CompositeName().add(add2.toString());
        }
        int n = add2.size() - 1;
        if (add.isEmpty() || add2.isEmpty() || add.get(0).equals("") || add2.get(n).equals("")) {
            return super.composeName(add, add2);
        }
        Name name = (Name)(Name)add2.clone();
        name.addAll(add);
        if (this.parentIsLdapCtx) {
            String concatNames = concatNames(name.get(n + 1), name.get(n));
            name.remove(n + 1);
            name.remove(n);
            name.add(n, concatNames);
            return name;
        }
        return name;
    }
    
    private String fullyQualifiedName(Name name) {
        return name.isEmpty() ? this.currentDN : this.fullyQualifiedName(name.get(0));
    }
    
    private String fullyQualifiedName(String s) {
        return concatNames(s, this.currentDN);
    }
    
    private static String concatNames(String str, String str2) {
        if (str == null || str.equals("")) {
            return str2;
        }
        if (str2 == null || str2.equals("")) {
            return str;
        }
        return str + "," + str2;
    }
    
    @Override
    protected Attributes c_getAttributes(Name remain, String[] returningAttributes, Continuation continuation) throws NamingException {
        continuation.setError(this, remain);
        SearchControls searchControls = new SearchControls();
        searchControls.setSearchScope(0);
        searchControls.setReturningAttributes(returningAttributes);
        try {
            LdapResult doSearchOnce = this.doSearchOnce(remain, "(objectClass=*)", searchControls, true);
            this.respCtls = doSearchOnce.resControls;
            if (doSearchOnce.status != 0) {
                this.processReturnCode(doSearchOnce, remain);
            }
            if (doSearchOnce.entries == null || doSearchOnce.entries.size() != 1) {
                return new BasicAttributes(true);
            }
            LdapEntry ldapEntry = (LdapEntry)doSearchOnce.entries.elementAt(0);
            Vector<Control> respCtls = ldapEntry.respCtls;
            if (respCtls != null) {
                appendVector(this.respCtls, respCtls);
            }
            this.setParents(ldapEntry.attributes, (Name)remain.clone());
            return ldapEntry.attributes;
        }
        catch (LdapReferralException e) {
            if (this.handleReferrals == 2) {
                throw continuation.fillInException(e);
            }
            while (true) {
                LdapReferralContext ldapReferralContext = (LdapReferralContext)e.getReferralContext(this.envprops, this.bindCtls);
                try {
                    return ldapReferralContext.getAttributes(remain, returningAttributes);
                }
                catch (LdapReferralException ex) {
                    e = ex;
                }
                finally {
                    ldapReferralContext.close();
                }
            }
        }
        catch (NamingException e2) {
            throw continuation.fillInException(e2);
        }
    }
    
    @Override
    protected void c_modifyAttributes(Name remain, int n, Attributes attributes, Continuation continuation) throws NamingException {
        continuation.setError(this, remain);
        try {
            this.ensureOpen();
            if (attributes == null || attributes.size() == 0) {
                return;
            }
            String fullyQualifiedName = this.fullyQualifiedName(remain);
            int convertToLdapModCode = convertToLdapModCode(n);
            int[] array = new int[attributes.size()];
            Attribute[] array2 = new Attribute[attributes.size()];
            NamingEnumeration<Object> all = attributes.getAll();
            for (int n2 = 0; n2 < array.length && all.hasMore(); ++n2) {
                array[n2] = convertToLdapModCode;
                array2[n2] = (Attribute)all.next();
            }
            LdapResult modify = this.clnt.modify(fullyQualifiedName, array, array2, this.reqCtls);
            this.respCtls = modify.resControls;
            if (modify.status != 0) {
                this.processReturnCode(modify, remain);
                return;
            }
        }
        catch (LdapReferralException e) {
            if (this.handleReferrals == 2) {
                throw continuation.fillInException(e);
            }
            while (true) {
                LdapReferralContext ldapReferralContext = (LdapReferralContext)e.getReferralContext(this.envprops, this.bindCtls);
                try {
                    ldapReferralContext.modifyAttributes(remain, n, attributes);
                    return;
                }
                catch (LdapReferralException ex) {
                    e = ex;
                }
                finally {
                    ldapReferralContext.close();
                }
            }
        }
        catch (IOException rootCause) {
            CommunicationException e2 = new CommunicationException(rootCause.getMessage());
            e2.setRootCause(rootCause);
            throw continuation.fillInException(e2);
        }
        catch (NamingException e3) {
            throw continuation.fillInException(e3);
        }
    }
    
    @Override
    protected void c_modifyAttributes(Name remain, ModificationItem[] array, Continuation continuation) throws NamingException {
        continuation.setError(this, remain);
        try {
            this.ensureOpen();
            if (array == null || array.length == 0) {
                return;
            }
            String fullyQualifiedName = this.fullyQualifiedName(remain);
            int[] array2 = new int[array.length];
            Attribute[] array3 = new Attribute[array.length];
            for (int i = 0; i < array2.length; ++i) {
                ModificationItem modificationItem = array[i];
                array2[i] = convertToLdapModCode(modificationItem.getModificationOp());
                array3[i] = modificationItem.getAttribute();
            }
            LdapResult modify = this.clnt.modify(fullyQualifiedName, array2, array3, this.reqCtls);
            this.respCtls = modify.resControls;
            if (modify.status != 0) {
                this.processReturnCode(modify, remain);
            }
        }
        catch (LdapReferralException e) {
            if (this.handleReferrals == 2) {
                throw continuation.fillInException(e);
            }
            while (true) {
                LdapReferralContext ldapReferralContext = (LdapReferralContext)e.getReferralContext(this.envprops, this.bindCtls);
                try {
                    ldapReferralContext.modifyAttributes(remain, array);
                    return;
                }
                catch (LdapReferralException ex) {
                    e = ex;
                }
                finally {
                    ldapReferralContext.close();
                }
            }
        }
        catch (IOException rootCause) {
            CommunicationException e2 = new CommunicationException(rootCause.getMessage());
            e2.setRootCause(rootCause);
            throw continuation.fillInException(e2);
        }
        catch (NamingException e3) {
            throw continuation.fillInException(e3);
        }
    }
    
    private static int convertToLdapModCode(int n) {
        switch (n) {
            case 1: {
                return 0;
            }
            case 2: {
                return 2;
            }
            case 3: {
                return 1;
            }
            default: {
                throw new IllegalArgumentException("Invalid modification code");
            }
        }
    }
    
    @Override
    protected DirContext c_getSchema(Name remain, Continuation continuation) throws NamingException {
        continuation.setError(this, remain);
        try {
            return this.getSchemaTree(remain);
        }
        catch (NamingException e) {
            throw continuation.fillInException(e);
        }
    }
    
    @Override
    protected DirContext c_getSchemaClassDefinition(Name remain, Continuation continuation) throws NamingException {
        continuation.setError(this, remain);
        try {
            Attribute value = this.c_getAttributes(remain, new String[] { "objectclass" }, continuation).get("objectclass");
            if (value == null || value.size() == 0) {
                return LdapCtx.EMPTY_SCHEMA;
            }
            Context context = (Context)this.c_getSchema(remain, continuation).lookup("ClassDefinition");
            HierMemDirCtx hierMemDirCtx = new HierMemDirCtx();
            NamingEnumeration<Object> all = value.getAll();
            while (all.hasMoreElements()) {
                String name = (String)all.nextElement();
                hierMemDirCtx.bind(name, (Object)context.lookup(name));
            }
            hierMemDirCtx.setReadOnly(new SchemaViolationException("Cannot update schema object"));
            return hierMemDirCtx;
        }
        catch (NamingException e) {
            throw continuation.fillInException(e);
        }
    }
    
    private DirContext getSchemaTree(Name name) throws NamingException {
        String schemaEntry = this.getSchemaEntry(name, true);
        DirContext buildSchemaTree = (DirContext)this.schemaTrees.get(schemaEntry);
        if (buildSchemaTree == null) {
            buildSchemaTree = this.buildSchemaTree(schemaEntry);
            this.schemaTrees.put(schemaEntry, buildSchemaTree);
            return buildSchemaTree;
        }
        return buildSchemaTree;
    }
    
    private DirContext buildSchemaTree(String str) throws NamingException {
        NamingEnumeration<SearchResult> searchAux = this.searchAux(new CompositeName().add(str), "(objectClass=subschema)", new SearchControls(0, 0L, 0, LdapCtx.SCHEMA_ATTRIBUTES, true, false), false, true, new Continuation());
        if (!searchAux.hasMore()) {
            throw new OperationNotSupportedException("Cannot get read subschemasubentry: " + str);
        }
        SearchResult searchResult = (SearchResult)searchAux.next();
        searchAux.close();
        Object object = searchResult.getObject();
        if (!(object instanceof LdapCtx)) {
            throw new NamingException("Cannot get schema object as DirContext: " + str);
        }
        return LdapSchemaCtx.createSchemaTree(this.envprops, str, (LdapCtx)object, searchResult.getAttributes(), this.netscapeSchemaBug);
    }
    
    private String getSchemaEntry(Name obj, boolean b) throws NamingException {
        SearchControls searchControls = new SearchControls(0, 0L, 0, new String[] { "subschemasubentry" }, false, false);
        NamingEnumeration<SearchResult> searchAux;
        try {
            searchAux = this.searchAux(obj, "objectclass=*", searchControls, b, true, new Continuation());
        }
        catch (NamingException ex) {
            if (!this.clnt.isLdapv3 && this.currentDN.length() == 0 && obj.isEmpty()) {
                throw new OperationNotSupportedException("Cannot get schema information from server");
            }
            throw ex;
        }
        if (!searchAux.hasMoreElements()) {
            throw new ConfigurationException("Requesting schema of nonexistent entry: " + obj);
        }
        SearchResult searchResult = (SearchResult)searchAux.next();
        searchAux.close();
        Attribute value = searchResult.getAttributes().get("subschemasubentry");
        if (value == null || value.size() < 0) {
            if (this.currentDN.length() == 0 && obj.isEmpty()) {
                throw new OperationNotSupportedException("Cannot read subschemasubentry of root DSE");
            }
            return this.getSchemaEntry(new CompositeName(), false);
        }
        else {
            return (String)(String)value.get();
        }
    }
    
    void setParents(Attributes attributes, Name name) throws NamingException {
        NamingEnumeration<Object> all = attributes.getAll();
        while (all.hasMore()) {
            ((LdapAttribute)all.next()).setParent(this, name);
        }
    }
    
    String getURL() {
        if (this.url == null) {
            this.url = LdapURL.toUrlString(this.hostname, this.port_number, this.currentDN, this.hasLdapsScheme);
        }
        return this.url;
    }
    
    protected NamingEnumeration<SearchResult> c_search(Name name, Attributes attributes, Continuation continuation) throws NamingException {
        return this.c_search(name, attributes, null, continuation);
    }
    
    @Override
    protected NamingEnumeration<SearchResult> c_search(Name remain, Attributes attrs, String[] returningAttributes, Continuation continuation) throws NamingException {
        SearchControls searchControls = new SearchControls();
        searchControls.setReturningAttributes(returningAttributes);
        String format;
        try {
            format = SearchFilter.format(attrs);
        }
        catch (NamingException e) {
            continuation.setError(this, remain);
            throw continuation.fillInException(e);
        }
        return this.c_search(remain, format, searchControls, continuation);
    }
    
    @Override
    protected NamingEnumeration<SearchResult> c_search(Name name, String s, SearchControls searchControls, Continuation continuation) throws NamingException {
        return this.searchAux(name, s, cloneSearchControls(searchControls), true, this.waitForReply, continuation);
    }
    
    @Override
    protected NamingEnumeration<SearchResult> c_search(Name remain, String expr, Object[] args, SearchControls searchControls, Continuation continuation) throws NamingException {
        String format;
        try {
            format = SearchFilter.format(expr, args);
        }
        catch (NamingException e) {
            continuation.setError(this, remain);
            throw continuation.fillInException(e);
        }
        return this.c_search(remain, format, searchControls, continuation);
    }
    
    NamingEnumeration<SearchResult> searchAux(Name remain, String s, SearchControls searchControls, boolean b, boolean b2, Continuation continuation) throws NamingException {
        LdapResult ldapResult = null;
        String[] array = new String[2];
        if (searchControls == null) {
            searchControls = new SearchControls();
        }
        String[] returningAttributes = searchControls.getReturningAttributes();
        if (searchControls.getReturningObjFlag() && returningAttributes != null) {
            boolean n = false;
            int i = returningAttributes.length - 1;
            while (i >= 0) {
                if (returningAttributes[i].equals("*")) {
                    n = true;
                    break;
                }
                --i;
            }
            if (n == 0) {
                String[] returningAttributes2 = new String[returningAttributes.length + Obj.JAVA_ATTRIBUTES.length];
                System.arraycopy(returningAttributes, 0, returningAttributes2, 0, returningAttributes.length);
                System.arraycopy(Obj.JAVA_ATTRIBUTES, 0, returningAttributes2, returningAttributes.length, Obj.JAVA_ATTRIBUTES.length);
                searchControls.setReturningAttributes(returningAttributes2);
            }
        }
        LdapCtx.SearchArgs o = new LdapCtx.SearchArgs(remain, s, searchControls, returningAttributes);
        continuation.setError(this, remain);
        try {
            if (searchToCompare(s, searchControls, array)) {
                ldapResult = this.compare(remain, array[0], array[1]);
                if (!ldapResult.compareToSearchResult(this.fullyQualifiedName(remain))) {
                    this.processReturnCode(ldapResult, remain);
                }
            }
            else {
                ldapResult = this.doSearch(remain, s, searchControls, b, b2);
                this.processReturnCode(ldapResult, remain);
            }
            return new LdapSearchEnumeration(this, ldapResult, this.fullyQualifiedName(remain), o, continuation);
        }
        catch (LdapReferralException e) {
            if (this.handleReferrals == 2) {
                throw continuation.fillInException(e);
            }
            while (true) {
                LdapReferralContext ldapReferralContext = (LdapReferralContext)e.getReferralContext(this.envprops, this.bindCtls);
                try {
                    return ldapReferralContext.search(remain, s, searchControls);
                }
                catch (LdapReferralException ex) {
                    e = ex;
                }
                finally {
                    ldapReferralContext.close();
                }
            }
        }
        catch (LimitExceededException namingException) {
            LdapSearchEnumeration o2 = new LdapSearchEnumeration(this, ldapResult, this.fullyQualifiedName(remain), o, continuation);
            o2.setNamingException(namingException);
            return o2;
        }
        catch (PartialResultException namingException2) {
            LdapSearchEnumeration o3 = new LdapSearchEnumeration(this, ldapResult, this.fullyQualifiedName(remain), o, continuation);
            o3.setNamingException(namingException2);
            return o3;
        }
        catch (IOException rootCause) {
            CommunicationException e2 = new CommunicationException(rootCause.getMessage());
            e2.setRootCause(rootCause);
            throw continuation.fillInException(e2);
        }
        catch (NamingException e3) {
            throw continuation.fillInException(e3);
        }
    }
    
    LdapResult getSearchReply(LdapClient ldapClient, LdapResult ldapResult) throws NamingException {
        if (this.clnt != ldapClient) {
            throw new CommunicationException("Context's connection changed; unable to continue enumeration");
        }
        try {
            return ldapClient.getSearchReply(this.batchSize, ldapResult, this.binaryAttrs);
        }
        catch (IOException rootCause) {
            CommunicationException ex = new CommunicationException(rootCause.getMessage());
            ex.setRootCause(rootCause);
            throw ex;
        }
    }
    
    private LdapResult doSearchOnce(Name name, String s, SearchControls searchControls, boolean b) throws NamingException {
        int batchSize = this.batchSize;
        this.batchSize = 2;
        LdapResult doSearch = this.doSearch(name, s, searchControls, b, true);
        this.batchSize = batchSize;
        return doSearch;
    }
    
    private LdapResult doSearch(Name name, String s, SearchControls searchControls, boolean b, boolean b2) throws NamingException {
        this.ensureOpen();
        try {
            int n;
            switch (searchControls.getSearchScope()) {
                case 0: {
                    n = 0;
                    break;
                }
                case 2: {
                    n = 2;
                    break;
                }
                default: {
                    n = 1;
                    break;
                }
            }
            String[] returningAttributes = searchControls.getReturningAttributes();
            if (returningAttributes != null && returningAttributes.length == 0) {
                returningAttributes = new String[] { "1.1" };
            }
            String s2 = b ? this.fullyQualifiedName(name) : (name.isEmpty() ? "" : name.get(0));
            int timeLimit = searchControls.getTimeLimit();
            int n2 = 0;
            if (timeLimit > 0) {
                n2 = timeLimit / 1000 + 1;
            }
            LdapResult search = this.clnt.search(s2, n, this.derefAliases, (int)searchControls.getCountLimit(), n2, !searchControls.getReturningObjFlag() && this.typesOnly, returningAttributes, s, this.batchSize, this.reqCtls, this.binaryAttrs, b2, this.replyQueueSize);
            this.respCtls = search.resControls;
            return search;
        }
        catch (IOException rootCause) {
            CommunicationException ex = new CommunicationException(rootCause.getMessage());
            ex.setRootCause(rootCause);
            throw ex;
        }
    }
    
    private static boolean searchToCompare(String s, SearchControls searchControls, String[] array) {
        if (searchControls.getSearchScope() != 0) {
            return false;
        }
        String[] returningAttributes = searchControls.getReturningAttributes();
        return returningAttributes != null && returningAttributes.length == 0 && filterToAssertion(s, array);
    }
    
    private static boolean filterToAssertion(String str, String[] array) {
        StringTokenizer stringTokenizer = new StringTokenizer(str, "=");
        if (stringTokenizer.countTokens() != 2) {
            return false;
        }
        array[0] = stringTokenizer.nextToken();
        array[1] = stringTokenizer.nextToken();
        if (array[1].indexOf(42) != -1) {
            return false;
        }
        boolean n = false;
        int length = array[1].length();
        if (array[0].charAt(0) == '(' && array[1].charAt(length - 1) == ')') {
            n = true;
        }
        else if (array[0].charAt(0) == '(' || array[1].charAt(length - 1) == ')') {
            return false;
        }
        if (new StringTokenizer(array[0], "()&|!=~><*", true).countTokens() != (n ? 2 : 1)) {
            return false;
        }
        if (new StringTokenizer(array[1], "()&|!=~><*", true).countTokens() != (n ? 2 : 1)) {
            return false;
        }
        if (!(n == 0)) {
            array[0] = array[0].substring(1);
            array[1] = array[1].substring(0, length - 1);
        }
        return true;
    }
    
    private LdapResult compare(Name name, String s, String s2) throws IOException, NamingException {
        this.ensureOpen();
        LdapResult compare = this.clnt.compare(this.fullyQualifiedName(name), s, s2, this.reqCtls);
        this.respCtls = compare.resControls;
        return compare;
    }
    
    private static SearchControls cloneSearchControls(SearchControls searchControls) {
        if (searchControls == null) {
            return null;
        }
        String[] returningAttributes = searchControls.getReturningAttributes();
        if (returningAttributes != null) {
            String[] array = new String[returningAttributes.length];
            System.arraycopy(returningAttributes, 0, array, 0, returningAttributes.length);
            returningAttributes = array;
        }
        return new SearchControls(searchControls.getSearchScope(), searchControls.getCountLimit(), searchControls.getTimeLimit(), returningAttributes, searchControls.getReturningObjFlag(), searchControls.getDerefLinkFlag());
    }
    
    @Override
    protected Hashtable<String, Object> p_getEnvironment() {
        return this.envprops;
    }
    
    @Override
    public Hashtable<String, Object> getEnvironment() throws NamingException {
        return (this.envprops == null) ? new Hashtable(5, 0.75f) : ((Hashtable)this.envprops.clone());
    }
    
    @Override
    public Object removeFromEnvironment(String s) throws NamingException {
        if (this.envprops == null || this.envprops.get(s) == null) {
            return null;
        }
        int n = -1;
        switch (s.hashCode()) {
            case 1186209693: {
                if (s.equals("java.naming.ldap.ref.separator")) {
                    n = 0;
                    break;
                }
                break;
            }
            case -1012288104: {
                if (s.equals("java.naming.ldap.typesOnly")) {
                    n = 1;
                    break;
                }
                break;
            }
            case -1907752796: {
                if (s.equals("java.naming.ldap.deleteRDN")) {
                    n = 2;
                    break;
                }
                break;
            }
            case 523906937: {
                if (s.equals("java.naming.ldap.derefAliases")) {
                    n = 3;
                    break;
                }
                break;
            }
            case -1690536287: {
                if (s.equals("java.naming.batchsize")) {
                    n = 4;
                    break;
                }
                break;
            }
            case -597571529: {
                if (s.equals("java.naming.ldap.referral.limit")) {
                    n = 5;
                    break;
                }
                break;
            }
            case 778463863: {
                if (s.equals("java.naming.referral")) {
                    n = 6;
                    break;
                }
                break;
            }
            case 704861003: {
                if (s.equals("java.naming.ldap.attributes.binary")) {
                    n = 7;
                    break;
                }
                break;
            }
            case 1514015564: {
                if (s.equals("com.sun.jndi.ldap.connect.timeout")) {
                    n = 8;
                    break;
                }
                break;
            }
            case 51576954: {
                if (s.equals("com.sun.jndi.ldap.read.timeout")) {
                    n = 9;
                    break;
                }
                break;
            }
            case 1515176491: {
                if (s.equals("com.sun.jndi.ldap.search.waitForReply")) {
                    n = 10;
                    break;
                }
                break;
            }
            case 1859284381: {
                if (s.equals("com.sun.jndi.ldap.search.replyQueueSize")) {
                    n = 11;
                    break;
                }
                break;
            }
            case -1793828884: {
                if (s.equals("java.naming.security.protocol")) {
                    n = 12;
                    break;
                }
                break;
            }
            case -2125978773: {
                if (s.equals("java.naming.ldap.version")) {
                    n = 13;
                    break;
                }
                break;
            }
            case -1264241948: {
                if (s.equals("java.naming.ldap.factory.socket")) {
                    n = 14;
                    break;
                }
                break;
            }
            case 520248524: {
                if (s.equals("java.naming.security.authentication")) {
                    n = 15;
                    break;
                }
                break;
            }
            case -986853030: {
                if (s.equals("java.naming.security.principal")) {
                    n = 16;
                    break;
                }
                break;
            }
            case -1273698008: {
                if (s.equals("java.naming.security.credentials")) {
                    n = 17;
                    break;
                }
                break;
            }
        }
        switch (n) {
            case 0: {
                this.addrEncodingSeparator = '#';
                break;
            }
            case 1: {
                this.typesOnly = false;
                break;
            }
            case 2: {
                this.deleteRDN = true;
                break;
            }
            case 3: {
                this.derefAliases = 3;
                break;
            }
            case 4: {
                this.batchSize = 1;
                break;
            }
            case 5: {
                this.referralHopLimit = 10;
                break;
            }
            case 6: {
                this.setReferralMode(null, true);
                break;
            }
            case 7: {
                this.setBinaryAttributes(null);
                break;
            }
            case 8: {
                this.connectTimeout = -1;
                break;
            }
            case 9: {
                this.readTimeout = -1;
                break;
            }
            case 10: {
                this.waitForReply = true;
                break;
            }
            case 11: {
                this.replyQueueSize = -1;
                break;
            }
            case 12: {
                this.closeConnection(false);
                if (this.useSsl && !this.hasLdapsScheme) {
                    this.useSsl = false;
                    this.url = null;
                    if (this.useDefaultPortNumber) {
                        this.port_number = 389;
                        break;
                    }
                    break;
                }
                else {
                    break;
                }
            }
            case 13:
            case 14: {
                this.closeConnection(false);
                break;
            }
            case 15:
            case 16:
            case 17: {
                this.sharable = false;
                break;
            }
        }
        this.envprops = (Hashtable)this.envprops.clone();
        return this.envprops.remove(s);
    }
    
    @Override
    public Object addToEnvironment(String key, Object o) throws NamingException {
        if (o == null) {
            return this.removeFromEnvironment(key);
        }
        int n = -1;
        switch (key.hashCode()) {
            case 1186209693: {
                if (key.equals("java.naming.ldap.ref.separator")) {
                    n = 0;
                    break;
                }
                break;
            }
            case -1012288104: {
                if (key.equals("java.naming.ldap.typesOnly")) {
                    n = 1;
                    break;
                }
                break;
            }
            case -1907752796: {
                if (key.equals("java.naming.ldap.deleteRDN")) {
                    n = 2;
                    break;
                }
                break;
            }
            case 523906937: {
                if (key.equals("java.naming.ldap.derefAliases")) {
                    n = 3;
                    break;
                }
                break;
            }
            case -1690536287: {
                if (key.equals("java.naming.batchsize")) {
                    n = 4;
                    break;
                }
                break;
            }
            case -597571529: {
                if (key.equals("java.naming.ldap.referral.limit")) {
                    n = 5;
                    break;
                }
                break;
            }
            case 778463863: {
                if (key.equals("java.naming.referral")) {
                    n = 6;
                    break;
                }
                break;
            }
            case 704861003: {
                if (key.equals("java.naming.ldap.attributes.binary")) {
                    n = 7;
                    break;
                }
                break;
            }
            case 1514015564: {
                if (key.equals("com.sun.jndi.ldap.connect.timeout")) {
                    n = 8;
                    break;
                }
                break;
            }
            case 51576954: {
                if (key.equals("com.sun.jndi.ldap.read.timeout")) {
                    n = 9;
                    break;
                }
                break;
            }
            case 1515176491: {
                if (key.equals("com.sun.jndi.ldap.search.waitForReply")) {
                    n = 10;
                    break;
                }
                break;
            }
            case 1859284381: {
                if (key.equals("com.sun.jndi.ldap.search.replyQueueSize")) {
                    n = 11;
                    break;
                }
                break;
            }
            case -1793828884: {
                if (key.equals("java.naming.security.protocol")) {
                    n = 12;
                    break;
                }
                break;
            }
            case -2125978773: {
                if (key.equals("java.naming.ldap.version")) {
                    n = 13;
                    break;
                }
                break;
            }
            case -1264241948: {
                if (key.equals("java.naming.ldap.factory.socket")) {
                    n = 14;
                    break;
                }
                break;
            }
            case 520248524: {
                if (key.equals("java.naming.security.authentication")) {
                    n = 15;
                    break;
                }
                break;
            }
            case -986853030: {
                if (key.equals("java.naming.security.principal")) {
                    n = 16;
                    break;
                }
                break;
            }
            case -1273698008: {
                if (key.equals("java.naming.security.credentials")) {
                    n = 17;
                    break;
                }
                break;
            }
        }
        switch (n) {
            case 0: {
                this.setRefSeparator((String)o);
                break;
            }
            case 1: {
                this.setTypesOnly((String)o);
                break;
            }
            case 2: {
                this.setDeleteRDN((String)o);
                break;
            }
            case 3: {
                this.setDerefAliases((String)o);
                break;
            }
            case 4: {
                this.setBatchSize((String)o);
                break;
            }
            case 5: {
                this.setReferralLimit((String)o);
                break;
            }
            case 6: {
                this.setReferralMode((String)o, true);
                break;
            }
            case 7: {
                this.setBinaryAttributes((String)o);
                break;
            }
            case 8: {
                this.setConnectTimeout((String)o);
                break;
            }
            case 9: {
                this.setReadTimeout((String)o);
                break;
            }
            case 10: {
                this.setWaitForReply((String)o);
                break;
            }
            case 11: {
                this.setReplyQueueSize((String)o);
                break;
            }
            case 12: {
                this.closeConnection(false);
                if ("ssl".equals(o)) {
                    this.useSsl = true;
                    this.url = null;
                    if (this.useDefaultPortNumber) {
                        this.port_number = 636;
                        break;
                    }
                    break;
                }
                else {
                    break;
                }
            }
            case 13:
            case 14: {
                this.closeConnection(false);
                break;
            }
            case 15:
            case 16:
            case 17: {
                this.sharable = false;
                break;
            }
        }
        this.envprops = ((this.envprops == null) ? new Hashtable(5, 0.75f) : ((Hashtable)this.envprops.clone()));
        return this.envprops.put(key, o);
    }
    
    void setProviderUrl(String value) {
        if (this.envprops != null) {
            this.envprops.put("java.naming.provider.url", value);
        }
    }
    
    void setDomainName(String value) {
        if (this.envprops != null) {
            this.envprops.put("com.sun.jndi.ldap.domainname", value);
        }
    }
    
    private void initEnv() throws NamingException {
        if (this.envprops == null) {
            this.setReferralMode(null, false);
            return;
        }
        this.setBatchSize((String)this.envprops.get("java.naming.batchsize"));
        this.setRefSeparator((String)this.envprops.get("java.naming.ldap.ref.separator"));
        this.setDeleteRDN((String)this.envprops.get("java.naming.ldap.deleteRDN"));
        this.setTypesOnly((String)this.envprops.get("java.naming.ldap.typesOnly"));
        this.setDerefAliases((String)this.envprops.get("java.naming.ldap.derefAliases"));
        this.setReferralLimit((String)this.envprops.get("java.naming.ldap.referral.limit"));
        this.setBinaryAttributes((String)this.envprops.get("java.naming.ldap.attributes.binary"));
        this.bindCtls = cloneControls((Control[])(Control[])this.envprops.get("java.naming.ldap.control.connect"));
        this.setReferralMode((String)this.envprops.get("java.naming.referral"), false);
        this.setConnectTimeout((String)this.envprops.get("com.sun.jndi.ldap.connect.timeout"));
        this.setReadTimeout((String)this.envprops.get("com.sun.jndi.ldap.read.timeout"));
        this.setWaitForReply((String)this.envprops.get("com.sun.jndi.ldap.search.waitForReply"));
        this.setReplyQueueSize((String)this.envprops.get("com.sun.jndi.ldap.search.replyQueueSize"));
        return;
    }
    
    private void setDeleteRDN(String s) {
        if (s != null && s.equalsIgnoreCase("false")) {
            this.deleteRDN = false;
        }
        else {
            this.deleteRDN = true;
        }
    }
    
    private void setTypesOnly(String s) {
        if (s != null && s.equalsIgnoreCase("true")) {
            this.typesOnly = true;
        }
        else {
            this.typesOnly = false;
        }
    }
    
    private void setBatchSize(String s) {
        if (s != null) {
            this.batchSize = Integer.parseInt(s);
        }
        else {
            this.batchSize = 1;
        }
    }
    
    private void setReferralMode(String s, boolean b) {
        if (s != null) {
            int n = -1;
            switch (s.hashCode()) {
                case -412873919: {
                    if (s.equals("follow-scheme")) {
                        n = 0;
                        break;
                    }
                    break;
                }
                case -1268958287: {
                    if (s.equals("follow")) {
                        n = 1;
                        break;
                    }
                    break;
                }
                case 110339814: {
                    if (s.equals("throw")) {
                        n = 2;
                        break;
                    }
                    break;
                }
                case -1190396462: {
                    if (s.equals("ignore")) {
                        n = 3;
                        break;
                    }
                    break;
                }
            }
            switch (n) {
                case 0: {
                    this.handleReferrals = 4;
                    break;
                }
                case 1: {
                    this.handleReferrals = 1;
                    break;
                }
                case 2: {
                    this.handleReferrals = 2;
                    break;
                }
                case 3: {
                    this.handleReferrals = 3;
                    break;
                }
                default: {
                    throw new IllegalArgumentException("Illegal value for java.naming.referral property.");
                }
            }
        }
        else {
            this.handleReferrals = 3;
        }
        if (this.handleReferrals == 3) {
            this.reqCtls = addControl(this.reqCtls, LdapCtx.manageReferralControl);
        }
        else if (b) {
            this.reqCtls = removeControl(this.reqCtls, LdapCtx.manageReferralControl);
        }
    }
    
    private void setDerefAliases(String s) {
        if (s != null) {
            int n = -1;
            switch (s.hashCode()) {
                case 104712844: {
                    if (s.equals("never")) {
                        n = 0;
                        break;
                    }
                    break;
                }
                case 1778217274: {
                    if (s.equals("searching")) {
                        n = 1;
                        break;
                    }
                    break;
                }
                case -853173367: {
                    if (s.equals("finding")) {
                        n = 2;
                        break;
                    }
                    break;
                }
                case -1414557169: {
                    if (s.equals("always")) {
                        n = 3;
                        break;
                    }
                    break;
                }
            }
            switch (n) {
                case 0: {
                    this.derefAliases = 0;
                    break;
                }
                case 1: {
                    this.derefAliases = 1;
                    break;
                }
                case 2: {
                    this.derefAliases = 2;
                    break;
                }
                case 3: {
                    this.derefAliases = 3;
                    break;
                }
                default: {
                    throw new IllegalArgumentException("Illegal value for java.naming.ldap.derefAliases property.");
                }
            }
        }
        else {
            this.derefAliases = 3;
        }
    }
    
    private void setRefSeparator(String s) throws NamingException {
        if (s != null && s.length() > 0) {
            this.addrEncodingSeparator = s.charAt(0);
        }
        else {
            this.addrEncodingSeparator = '#';
        }
    }
    
    private void setReferralLimit(String s) {
        if (s != null) {
            this.referralHopLimit = Integer.parseInt(s);
            if (this.referralHopLimit == 0) {
                this.referralHopLimit = Integer.MAX_VALUE;
            }
        }
        else {
            this.referralHopLimit = 10;
        }
    }
    
    void setHopCount(int hopCount) {
        this.hopCount = hopCount;
    }
    
    private void setConnectTimeout(String s) {
        if (s != null) {
            this.connectTimeout = Integer.parseInt(s);
        }
        else {
            this.connectTimeout = -1;
        }
    }
    
    private void setReplyQueueSize(String s) {
        if (s != null) {
            this.replyQueueSize = Integer.parseInt(s);
            if (this.replyQueueSize <= 0) {
                this.replyQueueSize = -1;
            }
        }
        else {
            this.replyQueueSize = -1;
        }
    }
    
    private void setWaitForReply(String s) {
        if (s != null && s.equalsIgnoreCase("false")) {
            this.waitForReply = false;
        }
        else {
            this.waitForReply = true;
        }
    }
    
    private void setReadTimeout(String s) {
        if (s != null) {
            this.readTimeout = Integer.parseInt(s);
        }
        else {
            this.readTimeout = -1;
        }
    }
    
    private static Vector<Vector<String>> extractURLs(String s) {
        int index;
        int initialCapacity;
        for (index = 0, initialCapacity = 0; (index = s.indexOf(10, index)) >= 0; ++index, ++initialCapacity) {}
        Vector<Object> vector = new Vector(initialCapacity);
        int beginIndex;
        int index2;
        for (beginIndex = s.indexOf(10) + 1; (index2 = s.indexOf(10, beginIndex)) >= 0; beginIndex = index2 + 1) {
            Vector obj = new Vector(1);
            obj.addElement(s.substring(beginIndex, index2));
            vector.addElement(obj);
        }
        Vector obj2 = new Vector(1);
        obj2.addElement(s.substring(beginIndex));
        vector.addElement(obj2);
        return vector;
    }
    
    private void setBinaryAttributes(String s) {
        if (s == null) {
            this.binaryAttrs = null;
        }
        else {
            this.binaryAttrs = new Hashtable(11, 0.75f);
            StringTokenizer stringTokenizer = new StringTokenizer(s.toLowerCase(Locale.ENGLISH), " ");
            while (stringTokenizer.hasMoreTokens()) {
                this.binaryAttrs.put(stringTokenizer.nextToken(), Boolean.TRUE);
            }
        }
    }
    
    @Override
    protected void finalize() {
        try {
            this.close();
        }
        catch (NamingException ex) {}
    }
    
    @Override
    public synchronized void close() throws NamingException {
        if (this.eventSupport != null) {
            this.eventSupport.cleanup();
            this.removeUnsolicited();
        }
        if (this.enumCount > 0) {
            this.closeRequested = true;
            return;
        }
        this.closeConnection(false);
        return;
    }
    
    @Override
    public void reconnect(Control[] array) throws NamingException {
        this.envprops = ((this.envprops == null) ? new Hashtable(5, 0.75f) : ((Hashtable)this.envprops.clone()));
        if (array == null) {
            this.envprops.remove("java.naming.ldap.control.connect");
            this.bindCtls = null;
        }
        else {
            this.envprops.put("java.naming.ldap.control.connect", this.bindCtls = cloneControls(array));
        }
        this.sharable = false;
        this.ensureOpen();
    }
    
    private static String getMechsAllowedToSendCredentials() {
        PrivilegedAction<String> action = ()/*  */ -> System.getProperty("jdk.jndi.ldap.mechsAllowedToSendCredentials");
        return (System.getSecurityManager() == null) ? ((String)action.run()) : ((String)AccessController.doPrivileged(action));
    }
    
    private static Set<String> getMechsFromPropertyValue(String s) {
        if (s == null || s.trim().isEmpty()) {
            return Collections.emptySet();
        }
        return Collections.unmodifiableSet((Set)Arrays.stream(s.split(",")).map(String::trim).filter(s2/* java.lang.String, */ -> !s2.trim().isEmpty()).collect(Collectors.toSet()));
    }
    
    private boolean isConnectionEncrypted() {
        return this.hasLdapsScheme || this.clnt.isUpgradedToStartTls();
    }
    
    private void ensureCanTransmitCredentials(String anotherString) throws NamingException {
        if ("none".equalsIgnoreCase(anotherString) || "anonymous".equalsIgnoreCase(anotherString)) {
            return;
        }
        String allowed_MECHS_SP_VALUE = (String)this.envprops.get("jdk.jndi.ldap.mechsAllowedToSendCredentials");
        boolean n = false;
        int n2 = LdapCtx.ALLOWED_MECHS_SP_VALUE != null || allowed_MECHS_SP_VALUE != null;
        if (!this.isConnectionEncrypted() && (this.contextSeenStartTlsEnabled || n2)) {
            if ("simple".equalsIgnoreCase(anotherString) && !this.envprops.containsKey("java.naming.security.principal")) {
                return;
            }
            if (allowed_MECHS_SP_VALUE == null) {
                n = true;
                allowed_MECHS_SP_VALUE = LdapCtx.ALLOWED_MECHS_SP_VALUE;
            }
            if ("all".equalsIgnoreCase(allowed_MECHS_SP_VALUE)) {
                return;
            }
            if (!(n ? LdapCtx.MECHS_ALLOWED_BY_SP : getMechsFromPropertyValue(allowed_MECHS_SP_VALUE)).contains(anotherString)) {
                throw new NamingException("Transmission of credentials over unsecured connection is not allowed");
            }
        }
        return;
    }
    
    private void ensureOpen() throws NamingException {
        this.ensureOpen(false);
    }
    
    private void ensureOpen(boolean b) throws NamingException {
        try {
            if (this.clnt == null) {
                this.schemaTrees = new Hashtable(11, 0.75f);
                this.connect(b);
            }
            else if (!this.sharable || b) {
                synchronized (this.clnt) {
                    if (!this.clnt.isLdapv3 || this.clnt.referenceCount > 1 || this.clnt.usingSaslStreams() || !this.clnt.conn.useable) {
                        this.closeConnection(false);
                    }
                }
                this.schemaTrees = new Hashtable(11, 0.75f);
                this.connect(b);
            }
        }
        finally {
            this.sharable = true;
        }
    }
    
    private void connect(boolean b) throws NamingException {
        String s = null;
        Object value = null;
        String anObject = null;
        String s2 = null;
        String s3 = null;
        String s4 = null;
        boolean equalsIgnoreCase = false;
        if (this.envprops != null) {
            s = (String)this.envprops.get("java.naming.security.principal");
            value = this.envprops.get("java.naming.security.credentials");
            s4 = (String)this.envprops.get("java.naming.ldap.version");
            anObject = (this.useSsl ? "ssl" : ((String)this.envprops.get("java.naming.security.protocol")));
            s2 = (String)this.envprops.get("java.naming.ldap.factory.socket");
            s3 = (String)this.envprops.get("java.naming.security.authentication");
            equalsIgnoreCase = "true".equalsIgnoreCase((String)this.envprops.get("com.sun.jndi.ldap.connect.pool"));
        }
        if (s2 == null) {
            s2 = ("ssl".equals(anObject) ? "javax.net.ssl.SSLSocketFactory" : null);
        }
        if (s3 == null) {
            s3 = ((s == null) ? "none" : "simple");
        }
        try {
            int n = this.clnt == null;
            int n2;
            if (!(n == 0)) {
                n2 = ((s4 != null) ? Integer.parseInt(s4) : 32);
                this.clnt = LdapClient.getInstance((boolean)equalsIgnoreCase, this.hostname, this.port_number, s2, this.connectTimeout, this.readTimeout, this.trace, n2, s3, this.bindCtls, anObject, s, value, this.envprops);
                this.contextSeenStartTlsEnabled |= this.clnt.isUpgradedToStartTls();
                if (this.clnt.authenticateCalled()) {
                    return;
                }
            }
            else {
                if (this.sharable && b) {
                    return;
                }
                n2 = 3;
            }
            LdapResult authenticate;
            synchronized (this.clnt.conn.startTlsLock) {
                this.ensureCanTransmitCredentials(s3);
                authenticate = this.clnt.authenticate((boolean)n, s, value, n2, s3, this.bindCtls, this.envprops);
            }
            this.respCtls = authenticate.resControls;
            if (authenticate.status != 0) {
                if (!(n == 0)) {
                    this.closeConnection(true);
                }
                this.processReturnCode(authenticate);
            }
        }
        catch (LdapReferralException ex) {
            if (this.handleReferrals == 2) {
                throw ex;
            }
            NamingException ex2 = null;
            String nextReferral;
            while ((nextReferral = ex.getNextReferral()) != null) {
                LdapURL o = new LdapURL(nextReferral);
                this.hostname = o.getHost();
                if (this.hostname != null && this.hostname.charAt(0) == '[') {
                    this.hostname = this.hostname.substring(1, this.hostname.length() - 1);
                }
                this.port_number = o.getPort();
                try {
                    this.connect(b);
                }
                catch (NamingException ex3) {
                    ex2 = ex3;
                }
            }
            if (ex2 != null) {
                throw (NamingException)(NamingException)ex2.fillInStackTrace();
            }
            throw new NamingException("Internal error processing referral during connection");
        }
    }
    
    private void closeConnection(boolean b) {
        this.removeUnsolicited();
        if (this.clnt != null) {
            this.clnt.close(this.reqCtls, b);
            this.clnt = null;
        }
    }
    
    synchronized void incEnumCount() {
        ++this.enumCount;
    }
    
    synchronized void decEnumCount() {
        --this.enumCount;
        if (this.enumCount == 0 && this.closeRequested) {
            try {
                this.close();
            }
            catch (NamingException ex) {}
        }
    }
    
    protected void processReturnCode(LdapResult ldapResult) throws NamingException {
        this.processReturnCode(ldapResult, null, this, null, this.envprops, null);
    }
    
    void processReturnCode(LdapResult ldapResult, Name name) throws NamingException {
        this.processReturnCode(ldapResult, new CompositeName().add(this.currentDN), this, name, this.envprops, this.fullyQualifiedName(name));
    }
    
    protected void processReturnCode(LdapResult ldapResult, Name resolvedName, Object resolvedObj, Name remainingName, Hashtable<?, ?> hashtable, String s) throws NamingException {
        String errorMessage = LdapClient.getErrorMessage(ldapResult.status, ldapResult.errorMessage);
        Object o = null;
        PartialResultException mapErrorCode;
        switch (ldapResult.status) {
            case 0: {
                if (ldapResult.referrals != null) {
                    String explanation = "Unprocessed Continuation Reference(s)";
                    if (this.handleReferrals == 3) {
                        mapErrorCode = new PartialResultException(explanation);
                        break;
                    }
                    int size = ldapResult.referrals.size();
                    String s2 = "Continuation Reference";
                    for (int i = 0; i < size; ++i) {
                        o = new LdapReferralException(resolvedName, resolvedObj, remainingName, s2, hashtable, s, this.handleReferrals, this.reqCtls);
                        o.setReferralInfo((Vector<String>)ldapResult.referrals.elementAt(i), true);
                        if (this.hopCount > 1) {
                            o.setHopCount(this.hopCount);
                        }
                        if (o != null) {
                            o.nextReferralEx = o;
                        }
                    }
                    ldapResult.referrals = null;
                    if (ldapResult.refEx == null) {
                        ldapResult.refEx = o;
                    }
                    else {
                        LdapReferralException ex;
                        for (ex = ldapResult.refEx; ex.nextReferralEx != null; ex = ex.nextReferralEx) {}
                        ex.nextReferralEx = o;
                    }
                    if (this.hopCount > this.referralHopLimit) {
                        LimitExceededException o2 = new LimitExceededException("Referral limit exceeded");
                        o2.setRootCause(o);
                        throw o2;
                    }
                }
                return;
            }
            case 10: {
                if (this.handleReferrals == 3) {
                    mapErrorCode = new PartialResultException(errorMessage);
                    break;
                }
                LdapReferralException rootCause = new LdapReferralException(resolvedName, resolvedObj, remainingName, errorMessage, hashtable, s, this.handleReferrals, this.reqCtls);
                Vector<String> vector;
                if (ldapResult.referrals == null) {
                    vector = null;
                }
                else if (this.handleReferrals == 4) {
                    vector = new Vector();
                    for (String e : (Vector)ldapResult.referrals.elementAt(0)) {
                        if (e.startsWith("ldap:")) {
                            vector.add(e);
                        }
                        else {
                            continue;
                        }
                    }
                    if (vector.isEmpty()) {
                        vector = null;
                    }
                }
                else {
                    vector = (Vector)ldapResult.referrals.elementAt(0);
                }
                rootCause.setReferralInfo(vector, false);
                if (this.hopCount > 1) {
                    rootCause.setHopCount(this.hopCount);
                }
                if (this.hopCount > this.referralHopLimit) {
                    LimitExceededException o3 = new LimitExceededException("Referral limit exceeded");
                    o3.setRootCause(rootCause);
                    mapErrorCode = o3;
                    break;
                }
                mapErrorCode = rootCause;
                break;
            }
            case 9: {
                if (this.handleReferrals == 3) {
                    mapErrorCode = new PartialResultException(errorMessage);
                    break;
                }
                if (ldapResult.errorMessage != null && !ldapResult.errorMessage.equals("")) {
                    ldapResult.referrals = extractURLs(ldapResult.errorMessage);
                    LdapReferralException ex2 = new LdapReferralException(resolvedName, resolvedObj, remainingName, errorMessage, hashtable, s, this.handleReferrals, this.reqCtls);
                    if (this.hopCount > 1) {
                        ex2.setHopCount(this.hopCount);
                    }
                    if ((ldapResult.entries == null || ldapResult.entries.isEmpty()) && ldapResult.referrals != null && ldapResult.referrals.size() == 1) {
                        ex2.setReferralInfo(ldapResult.referrals, false);
                        if (this.hopCount > this.referralHopLimit) {
                            LimitExceededException ex3 = new LimitExceededException("Referral limit exceeded");
                            ex3.setRootCause(ex2);
                            mapErrorCode = ex3;
                            break;
                        }
                        mapErrorCode = ex2;
                        break;
                    }
                    else {
                        ex2.setReferralInfo(ldapResult.referrals, true);
                        ldapResult.refEx = ex2;
                        return;
                    }
                }
                else {
                    mapErrorCode = new PartialResultException(errorMessage);
                    break;
                }
            }
            case 34:
            case 64: {
                if (remainingName != null) {
                    mapErrorCode = new InvalidNameException(remainingName.toString() + ": " + errorMessage);
                    break;
                }
                mapErrorCode = new InvalidNameException(errorMessage);
                break;
            }
            default: {
                mapErrorCode = mapErrorCode(ldapResult.status, ldapResult.errorMessage);
                break;
            }
        }
        mapErrorCode.setResolvedName(resolvedName);
        mapErrorCode.setResolvedObj(resolvedObj);
        mapErrorCode.setRemainingName(remainingName);
        throw mapErrorCode;
    }
    
    public static NamingException mapErrorCode(int n, String s) {
        if (n == 0) {
            return null;
        }
        String errorMessage = LdapClient.getErrorMessage(n, s);
        NamingException ex;
        switch (n) {
            case 36: {
                ex = new NamingException(errorMessage);
                break;
            }
            case 33: {
                ex = new NamingException(errorMessage);
                break;
            }
            case 20: {
                ex = new AttributeInUseException(errorMessage);
                break;
            }
            case 7:
            case 8:
            case 13:
            case 48: {
                ex = new AuthenticationNotSupportedException(errorMessage);
                break;
            }
            case 68: {
                ex = new NameAlreadyBoundException(errorMessage);
                break;
            }
            case 14:
            case 49: {
                ex = new AuthenticationException(errorMessage);
                break;
            }
            case 18: {
                ex = new InvalidSearchFilterException(errorMessage);
                break;
            }
            case 50: {
                ex = new NoPermissionException(errorMessage);
                break;
            }
            case 19:
            case 21: {
                ex = new InvalidAttributeValueException(errorMessage);
                break;
            }
            case 54: {
                ex = new NamingException(errorMessage);
                break;
            }
            case 16: {
                ex = new NoSuchAttributeException(errorMessage);
                break;
            }
            case 32: {
                ex = new NameNotFoundException(errorMessage);
                break;
            }
            case 65:
            case 67:
            case 69: {
                ex = new SchemaViolationException(errorMessage);
                break;
            }
            case 66: {
                ex = new ContextNotEmptyException(errorMessage);
                break;
            }
            case 1: {
                ex = new NamingException(errorMessage);
                break;
            }
            case 80: {
                ex = new NamingException(errorMessage);
                break;
            }
            case 2: {
                ex = new CommunicationException(errorMessage);
                break;
            }
            case 4: {
                ex = new SizeLimitExceededException(errorMessage);
                break;
            }
            case 3: {
                ex = new TimeLimitExceededException(errorMessage);
                break;
            }
            case 12: {
                ex = new OperationNotSupportedException(errorMessage);
                break;
            }
            case 51:
            case 52: {
                ex = new ServiceUnavailableException(errorMessage);
                break;
            }
            case 17: {
                ex = new InvalidAttributeIdentifierException(errorMessage);
                break;
            }
            case 53: {
                ex = new OperationNotSupportedException(errorMessage);
                break;
            }
            case 5:
            case 6:
            case 35: {
                ex = new NamingException(errorMessage);
                break;
            }
            case 11: {
                ex = new LimitExceededException(errorMessage);
                break;
            }
            case 10: {
                ex = new NamingException(errorMessage);
                break;
            }
            case 9: {
                ex = new NamingException(errorMessage);
                break;
            }
            case 34:
            case 64: {
                ex = new InvalidNameException(errorMessage);
                break;
            }
            default: {
                ex = new NamingException(errorMessage);
                break;
            }
        }
        return ex;
    }
    
    @Override
    public ExtendedResponse extendedOperation(ExtendedRequest extendedRequest) throws NamingException {
        int equals = extendedRequest.getID().equals("1.3.6.1.4.1.1466.20037");
        this.ensureOpen((boolean)equals);
        try {
            LdapResult extendedOp = this.clnt.extendedOp(extendedRequest.getID(), extendedRequest.getEncodedValue(), this.reqCtls, (boolean)equals);
            this.respCtls = extendedOp.resControls;
            if (extendedOp.status != 0) {
                this.processReturnCode(extendedOp, new CompositeName());
            }
            ExtendedResponse extendedResponse = extendedRequest.createExtendedResponse(extendedOp.extensionId, extendedOp.extensionValue, 0, (extendedOp.extensionValue == null) ? 0 : extendedOp.extensionValue.length);
            if ((extendedResponse instanceof StartTlsResponseImpl)) {
                ((StartTlsResponseImpl)extendedResponse).setConnection(this.clnt.conn, (String)(String)((this.envprops != null) ? this.envprops.get("com.sun.jndi.ldap.domainname") : null));
                this.contextSeenStartTlsEnabled |= equals;
                return extendedResponse;
            }
            return extendedResponse;
        }
        catch (LdapReferralException ex) {
            if (this.handleReferrals == 2) {
                throw ex;
            }
            while (true) {
                LdapReferralContext ldapReferralContext = (LdapReferralContext)ex.getReferralContext(this.envprops, this.bindCtls);
                try {
                    return ldapReferralContext.extendedOperation(extendedRequest);
                }
                catch (LdapReferralException ex2) {
                    ex = ex2;
                }
                finally {
                    ldapReferralContext.close();
                }
            }
        }
        catch (IOException rootCause) {
            CommunicationException ex3 = new CommunicationException(rootCause.getMessage());
            ex3.setRootCause(rootCause);
            throw ex3;
        }
    }
    
    @Override
    public void setRequestControls(Control[] array) throws NamingException {
        if (this.handleReferrals == 3) {
            this.reqCtls = addControl(array, LdapCtx.manageReferralControl);
        }
        else {
            this.reqCtls = cloneControls(array);
        }
    }
    
    @Override
    public Control[] getRequestControls() throws NamingException {
        return cloneControls(this.reqCtls);
    }
    
    @Override
    public Control[] getConnectControls() throws NamingException {
        return cloneControls(this.bindCtls);
    }
    
    @Override
    public Control[] getResponseControls() throws NamingException {
        return (this.respCtls != null) ? this.convertControls(this.respCtls) : null;
    }
    
    Control[] convertControls(Vector<Control> vector) throws NamingException {
        int size = vector.size();
        if (size == 0) {
            return null;
        }
        Control[] array = new Control[size];
        for (int i = 0; i < size; ++i) {
            array[i] = LdapCtx.myResponseControlFactory.getControlInstance((Control)vector.elementAt(i));
            if (array[i] == null) {
                array[i] = ControlFactory.getControlInstance((Control)vector.elementAt(i), this, this.envprops);
            }
        }
        return array;
    }
    
    private static Control[] addControl(Control[] array, Control control) {
        if (array == null) {
            return new Control[] { control };
        }
        if (findControl(array, control) != -1) {
            return array;
        }
        Control[] array2 = new Control[array.length + 1];
        System.arraycopy(array, 0, array2, 0, array.length);
        array2[array.length] = control;
        return array2;
    }
    
    private static int findControl(Control[] array, Control control) {
        int i = 0;
        while (i < array.length) {
            if (array[i] == control) {
                return i;
            }
            ++i;
        }
        return -1;
    }
    
    private static Control[] removeControl(Control[] array, Control control) {
        if (array == null) {
            return null;
        }
        int control2 = findControl(array, control);
        if (control2 == -1) {
            return array;
        }
        Control[] array2 = new Control[array.length - 1];
        System.arraycopy(array, 0, array2, 0, control2);
        System.arraycopy(array, control2 + 1, array2, control2, array.length - control2 - 1);
        return array2;
    }
    
    private static Control[] cloneControls(Control[] array) {
        if (array == null) {
            return null;
        }
        Control[] array2 = new Control[array.length];
        System.arraycopy(array, 0, array2, 0, array.length);
        return array2;
    }
    
    @Override
    public void addNamingListener(Name name, int n, NamingListener namingListener) throws NamingException {
        this.addNamingListener(getTargetName(name), n, namingListener);
    }
    
    @Override
    public void addNamingListener(String n, int n2, NamingListener namingListener) throws NamingException {
        if (this.eventSupport == null) {
            this.eventSupport = new EventSupport(this);
        }
        this.eventSupport.addNamingListener(getTargetName(new CompositeName(n)), n2, namingListener);
        if (namingListener instanceof UnsolicitedNotificationListener && !this.unsolicited) {
            this.addUnsolicited();
        }
    }
    
    @Override
    public void removeNamingListener(NamingListener namingListener) throws NamingException {
        if (this.eventSupport == null) {
            return;
        }
        this.eventSupport.removeNamingListener(namingListener);
        if (namingListener instanceof UnsolicitedNotificationListener && !this.eventSupport.hasUnsolicited()) {
            this.removeUnsolicited();
        }
        return;
    }
    
    @Override
    public void addNamingListener(String n, String s, SearchControls searchControls, NamingListener namingListener) throws NamingException {
        if (this.eventSupport == null) {
            this.eventSupport = new EventSupport(this);
        }
        this.eventSupport.addNamingListener(getTargetName(new CompositeName(n)), s, cloneSearchControls(searchControls), namingListener);
        if (namingListener instanceof UnsolicitedNotificationListener && !this.unsolicited) {
            this.addUnsolicited();
        }
    }
    
    @Override
    public void addNamingListener(Name name, String s, SearchControls searchControls, NamingListener namingListener) throws NamingException {
        this.addNamingListener(getTargetName(name), s, searchControls, namingListener);
    }
    
    @Override
    public void addNamingListener(Name name, String s, Object[] array, SearchControls searchControls, NamingListener namingListener) throws NamingException {
        this.addNamingListener(getTargetName(name), s, array, searchControls, namingListener);
    }
    
    @Override
    public void addNamingListener(String n, String expr, Object[] args, SearchControls searchControls, NamingListener namingListener) throws NamingException {
        this.addNamingListener(getTargetName(new CompositeName(n)), SearchFilter.format(expr, args), searchControls, namingListener);
    }
    
    @Override
    public boolean targetMustExist() {
        return true;
    }
    
    private static String getTargetName(Name obj) throws NamingException {
        if ((obj instanceof CompositeName)) {
            if (obj.size() > 1) {
                throw new InvalidNameException("Target cannot span multiple namespaces: " + obj);
            }
            if (obj.isEmpty()) {
                return "";
            }
            return obj.get(0);
        }
        else {
            return obj.toString();
        }
    }
    
    private void addUnsolicited() throws NamingException {
        this.ensureOpen();
        synchronized (this.eventSupport) {
            this.clnt.addUnsolicited(this);
            this.unsolicited = true;
        }
    }
    
    private void removeUnsolicited() {
        if (this.eventSupport == null) {
            return;
        }
        synchronized (this.eventSupport) {
            if (this.unsolicited && this.clnt != null) {
                this.clnt.removeUnsolicited(this);
            }
            this.unsolicited = false;
        }
        return;
    }
    
    void fireUnsolicited(Object o) {
        synchronized (this.eventSupport) {
            if (this.unsolicited) {
                this.eventSupport.fireUnsolicited(o);
                if ((o instanceof NamingException)) {
                    this.unsolicited = false;
                }
            }
        }
    }
    
    static {
        SCHEMA_ATTRIBUTES = new String[] { "objectClasses", "attributeTypes", "matchingRules", "ldapSyntaxes" };
        ALLOWED_MECHS_SP_VALUE = getMechsAllowedToSendCredentials();
        MECHS_ALLOWED_BY_SP = getMechsFromPropertyValue(LdapCtx.ALLOWED_MECHS_SP_VALUE);
        parser = new LdapNameParser();
        myResponseControlFactory = new DefaultResponseControlFactory();
        manageReferralControl = new ManageReferralControl(false);
        (EMPTY_SCHEMA = new HierMemDirCtx()).setReadOnly(new SchemaViolationException("Cannot update schema object"));
    }
    
    static final class SearchArgs
    {
        Name name;
        String filter;
        SearchControls cons;
        String[] reqAttrs;
        
        SearchArgs(Name name, String filter, SearchControls cons, String[] reqAttrs) {
            this.name = name;
            this.filter = filter;
            this.cons = cons;
            this.reqAttrs = reqAttrs;
        }
    }
}
